//## begin module%1.4%.codegen_version preserve=yes
//   Read the documentation to learn more about C++ code generator
//   versioning.
//## end module%1.4%.codegen_version

//## begin module%418011BB031C.cm preserve=no
//	  %X% %Q% %Z% %W%
//## end module%418011BB031C.cm

//## begin module%418011BB031C.cp preserve=no
//## end module%418011BB031C.cp

//## Module: Chain%418011BB031C; Package body
//## Subsystem: XSFit::MCMC%41E6AA18032C
//## Source file: /local/data/xerxes1/gordon/headas/Xspec/src/XSFit/MCMC/Chain.cxx

//## begin module%418011BB031C.additionalIncludes preserve=no
//## end module%418011BB031C.additionalIncludes

//## begin module%418011BB031C.includes preserve=yes
#include <XSContainer.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/StatMethod.h>
#include <XSstreams.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <XSUtil/Numerics/RandomGenerator.h>
#include <XSUtil/Signals/SignalHandler.h>
#include <cmath>
#include <sstream>
#include <utility>
#include <cstdlib>
//## end module%418011BB031C.includes

// ChainIO
#include <XSFit/MCMC/ChainIO.h>
// Chain
#include <XSFit/MCMC/Chain.h>


//## begin module%418011BB031C.declarations preserve=no
//## end module%418011BB031C.declarations

//## begin module%418011BB031C.additionalDeclarations preserve=yes
using namespace XSContainer;
//## end module%418011BB031C.additionalDeclarations


// Class Chain::ChainError 

Chain::ChainError::ChainError (const string& msg)
  //## begin Chain::ChainError::ChainError%41813DFD0219.hasinit preserve=yes
  : YellowAlert("Chain Error: ")
  //## end Chain::ChainError::ChainError%41813DFD0219.hasinit
  //## begin Chain::ChainError::ChainError%41813DFD0219.initialization preserve=yes
  //## end Chain::ChainError::ChainError%41813DFD0219.initialization
{
  //## begin Chain::ChainError::ChainError%41813DFD0219.body preserve=yes
  tcerr << msg << std::endl;
  //## end Chain::ChainError::ChainError%41813DFD0219.body
}


// Class Chain::ParamID 

// Class Chain::ChainInterrupt 

Chain::ChainInterrupt::ChainInterrupt ()
  //## begin Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.hasinit preserve=yes
  : YellowAlert()
  //## end Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.hasinit
  //## begin Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.initialization preserve=yes
  //## end Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.initialization
{
  //## begin Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.body preserve=yes
  //## end Chain::ChainInterrupt::ChainInterrupt%4601442C00EA.body
}


// Class Chain 

Chain::Chain (ChainIO* pIO, const string& fileName, size_t burnLength, size_t length, bool rand)
  //## begin Chain::Chain%41811665000F.hasinit preserve=no
  //## end Chain::Chain%41811665000F.hasinit
  //## begin Chain::Chain%41811665000F.initialization preserve=yes
  : m_fileName(fileName),
    m_burnLength(burnLength),
    m_length(length),
    m_rand(rand),
    m_width(0),
    m_paramIDs(),
    m_statistic(),
    m_tempering(),
    m_IO(pIO)
  //## end Chain::Chain%41811665000F.initialization
{
  //## begin Chain::Chain%41811665000F.body preserve=yes
   // Nothing must throw in here, otherwise additional precautions
   // must be taken in chain run handler for deletion of m_IO.
   m_IO->setParent(this);
  //## end Chain::Chain%41811665000F.body
}

Chain::Chain (ChainIO* pIO, const string& fileName)
  //## begin Chain::Chain%41824C1F0153.hasinit preserve=yes
  : m_fileName(fileName),
    m_burnLength(0),
    m_length(0),
    m_rand(false),
    m_width(0),
    m_paramIDs(),
    m_statistic(),
    m_tempering(),
    m_IO(pIO)
  //## end Chain::Chain%41824C1F0153.hasinit
  //## begin Chain::Chain%41824C1F0153.initialization preserve=yes
  //## end Chain::Chain%41824C1F0153.initialization
{
  //## begin Chain::Chain%41824C1F0153.body preserve=yes
  m_IO->setParent(this);
  // This may throw:
  m_IO->readFileInfo();
  //## end Chain::Chain%41824C1F0153.body
}


Chain::~Chain()
{
  //## begin Chain::~Chain%4180176A01EC_dest.body preserve=yes
   delete m_IO;
  //## end Chain::~Chain%4180176A01EC_dest.body
}



//## Other Operations (implementation)
void Chain::run (const size_t appendLength, const Real temperature)
{
  //## begin Chain::run%418119F4027E.body preserve=yes
  using namespace std;
  // ASSUME temperature has already been checked != 0.0
  if (!m_length)
  {
     string msg("Chain length is zero - please set it.");
     throw ChainError(msg);
  }

  bool isTempChange = false;
  const Real EPS = 1.0e-15;
  if (m_tempering.size())
  {
     // Should only get in here for append.
     Real lastTemp = m_tempering.back().second;
     isTempChange = (abs(temperature-lastTemp)/lastTemp > EPS);
     if (isTempChange && !m_IO->allowsTemper())
     {
        string errMsg("Cannot append to old-style chain file with a different temperature.");
        errMsg += "\n   File does not have a temperature values field.";
        throw ChainError(errMsg);
     }
     if (isTempChange  && !m_IO->checkTemperField(temperature))
     {
        string errMsg("Cannot append another temperature to ");
        errMsg += m_fileName + "\n     Tempering info field is full.";
        throw ChainError(errMsg);
     }
  }
    
  const bool wasFitStillValid = fit->isStillValid();
  const map<int,ModParam*>& varPars = fit->variableParameters();
  map<int,ModParam*>::const_iterator itVp = varPars.begin();
  map<int,ModParam*>::const_iterator itVpEnd = varPars.end();
  
  // Until this is implemented for response params...
  while (itVp != itVpEnd)
  {
     if (dynamic_cast<ResponseParam*>(itVp->second))
     {
        string errMsg("Chain runs are not yet implemented for response parameters.");
        throw ChainError(errMsg);
     }
     ++itVp;
  }
  itVp = varPars.begin();
  
  vector<Real> originalParamValues;
  m_paramIDs.clear();
  const size_t nPars = varPars.size();
  if (!nPars)
  {
     throw ChainError("No variable parameters");
  }
  // This first call with flag=true will only perform Randomizer
  // initialization.  Since it may throw, let's get it out of the
  // way before we start messing around with files.
  fit->randomizeForChain(true);
  
  m_paramIDs.resize(nPars);
  for (size_t i=0; i<nPars; ++i, ++itVp)
  {
     const ModParam* param = itVp->second;
     ParamID& parID = m_paramIDs[i];
     parID.modName = param->modelName();
     parID.parName = param->name();
     parID.index = param->index();
     parID.units = param->unit();
     originalParamValues.push_back(param->value());
  }
  
  m_width = nPars + 1;
  const size_t origLength = appendLength ? m_length : 0;
  if (appendLength)
  {
     RealArray startingVals;
     m_IO->appendFile(startingVals);
     if (varPars.size() != startingVals.size())
     {
        string errMsg("Pars in ");
        errMsg += m_fileName;
        errMsg += string(" do not match fit variable parameters.");
        throw ChainError(errMsg);
     }
     size_t iPar=0;
     itVp = varPars.begin();
     itVpEnd = varPars.end();
     while (itVp != itVpEnd)
     {
        itVp->second->setValue(startingVals[iPar],'v');
        ++itVp, ++iPar;
     }
     fit->isStillValid(false);
     fit->statMethod()->perform(fit);  
  }
  else
  {
     // Creating a new chain file.
     m_tempering.push_back(make_pair(size_t(1), temperature));
     m_IO->createFile(); 
     m_IO->writeFileInfo(); 
  }
  
  size_t i=1;
  // Initialize the chain with the first point (but not
  // if appending).
  try
  {
     if (!appendLength)
     {
        if (m_rand)
        {
           fit->randomizeForChain();
           fit->statMethod()->perform(fit);
        }  
        if (!m_burnLength)
           m_IO->writePoint();
        ++i;
     }
  }
  catch (YellowAlert&)
  {
     cleanupRun(originalParamValues);
     fit->isStillValid(wasFitStillValid);
     throw;
  }
  
  std::vector<Real> saveParamValues(nPars);
  size_t totalLength = appendLength ? appendLength : m_length + m_burnLength;
  const Numerics::DefaultRandomGenerator& randGen = 
        Numerics::DefaultRandomGenerator::instance();
        
  // If SIGINT handler hasn't been registered, intHandler will be 0
  // and it won't allow ctrl-C breaking from the loop.      
  const SIGINT_Handler *intHandler = dynamic_cast<const SIGINT_Handler*>
                (SignalHandler::instance()->getHandler(SIGINT));
  try
  {
     // Not resetting verbosity for this since we don't want to affect
     // output in the nested functions below.
     const bool doChatter = (tpout.maxChatter() >= 14);
     
     while (i<=totalLength)
     {
        if (intHandler && intHandler->interrupted())
        {
           throw ChainInterrupt();
        }
        itVp = varPars.begin();
        for (size_t j=0; j<nPars; ++j, ++itVp)
        {
           saveParamValues[j] = itVp->second->value();
        }
        Real saveStat = fit->statistic();
        fit->randomizeForChain();
        
	// for high chatter tell user values we are trying
	if ( doChatter )
	{
	   tcout << "Trying parameters for chain: ";
	   itVp = varPars.begin();
           itVpEnd = varPars.end();
	   while (itVp != itVpEnd)
	   {
              tcout << itVp->second->value() << " ";
              ++itVp;
	   }
	   tcout << std::endl;
	}
                
        fit->statMethod()->perform(fit);
        
	if ( doChatter ) 
	{
	   tcout << "Old -> New statistic: " << saveStat
	         << " -> " << fit->statistic()
	         << " (delta="
	         << saveStat-fit->statistic() << ")"
	         << std::endl;
	}
                
        bool includeNewParams = false;
        if (fit->statistic() <= saveStat)
        {
           includeNewParams = true;
           if (doChatter)
	      tcout << "New parameters were included (statistic was lower)\n"
		    << std::endl;
        }
        else
        {
           float randNumber=0.;
           randGen.getRandom(&randNumber, 1); 
           Real weightedDiff(exp(.5*(saveStat-fit->statistic())/temperature));
           if (randNumber < weightedDiff)
              includeNewParams = true;
	   if( doChatter )
	   {
	     tcout << "Criterion for inclusion in chain: "
		   << randNumber << " < " << weightedDiff << std::endl;

	     if(includeNewParams)
	       tcout << "=> parameters were included (statistic was higher)\n"
		     << std::endl;
	     else
	       tcout << "=> parameters NOT included (statistic was higher)\n"
		     << std::endl;
	   }
        }
        fit->reportChainAcceptance(includeNewParams);
        
        if (!includeNewParams)
        {
           itVp = varPars.begin();
           for (size_t j=0; j<nPars; ++j, ++itVp)
           {
              itVp->second->setValue(saveParamValues[j],'v');
           }
           fit->statMethod()->statistic(saveStat);
        }
        if (i > m_burnLength || appendLength)
           m_IO->writePoint();
        ++i;
     }
     if (appendLength)
     {
        m_length += appendLength;
        m_IO->doCloseFile();
        m_IO->doOpenForWrite();
        m_IO->adjustLengthInfo();
        if (isTempChange)
        {
           m_tempering.push_back(make_pair(origLength+(size_t)1, temperature));
           m_IO->adjustTemperInfo();
        }
     }
  }
  catch (YellowAlert&)
  {
     // The -1 comes from the fact that i would have been incremented
     // one more time after the last point was written.
     if (appendLength)
     {
        m_length = origLength + i-1;
        m_IO->doCloseFile();
        m_IO->doOpenForWrite();
        if (m_length != origLength && isTempChange)
        {
           m_tempering.push_back(make_pair(origLength+(size_t)1, temperature));
           m_IO->adjustTemperInfo();
        }
        tcerr << "\n*** Warning: Chain append run has been interrupted." << std::endl;
     }
     else
     {
        m_length = (i-1 > m_burnLength) ? i-1-m_burnLength : 0;
        tcerr << "\n*** Warning: Chain run has been interrupted.\n"
             <<  "    The chain will not be loaded." << std::endl;
     }
     m_IO->adjustLengthInfo();     
     cleanupRun(originalParamValues);
     fit->isStillValid(wasFitStillValid);
     // Even if exception is coming from something other than a
     // Ctrl-C break, we'll treat it the same way in this context.
     throw ChainInterrupt();
  }
  cleanupRun(originalParamValues);
  fit->isStillValid(wasFitStillValid);
  //## end Chain::run%418119F4027E.body
}

void Chain::cleanupRun (const std::vector<Real>& origPars)
{
  //## begin Chain::cleanupRun%4190EC2501B1.body preserve=yes
  m_IO->doCloseFile();
  std::map<int,ModParam*>::const_iterator itVp = 
                fit->variableParameters().begin();
  size_t nPars = fit->variableParameters().size();
  for (size_t i=0; i<nPars; ++i, ++itVp)
  {
     itVp->second->setValue(origPars[i],'v');
  }
  fit->statMethod()->perform(fit);
  //## end Chain::cleanupRun%4190EC2501B1.body
}

void Chain::openForReadPoints () const
{
  //## begin Chain::openForReadPoints%4194E2DD0211.body preserve=yes
  m_IO->doOpenForReadPoints();
  //## end Chain::openForReadPoints%4194E2DD0211.body
}

void Chain::closeFile () const
{
  //## begin Chain::closeFile%4194E31601FF.body preserve=yes
  m_IO->doCloseFile();
  //## end Chain::closeFile%4194E31601FF.body
}

void Chain::readPoint (std::vector<Real>& chainVals) const
{
  //## begin Chain::readPoint%4194E37B03B3.body preserve=yes
  m_IO->doReadPoint(chainVals);
  //## end Chain::readPoint%4194E37B03B3.body
}

void Chain::calcStats (size_t iPar, Real& mean, Real& variance, size_t& nRepeats)
{
  //## begin Chain::calcStats%4198DB3602EE.body preserve=yes
  // Assumes iPar is already verified to be >= 0.
  if (iPar >= m_width-1)
  {
     std::ostringstream msg;
     msg << "Stored chain width value is not consistent with parameter labels in file "
         << m_fileName;
     throw RedAlert(msg.str());
  }
  RealArray parVals(.0,m_length);
  Real total = .0;
  m_IO->doOpenForReadPoints();
  // This setting of prevPoint is not likely to match initial point.
  std::vector<Real> prevPoint(m_width,-9.876e99);
  nRepeats = 0;
  std::vector<Real> tmp;
  for (size_t i=0; i<m_length; ++i)
  {
     m_IO->doReadPoint(tmp);
     if (checkForRepeat(prevPoint, tmp))
        ++nRepeats;
     prevPoint = tmp;
     parVals[i] = tmp[iPar];
     total += parVals[i];
  }
  m_IO->doCloseFile();
  mean = total/m_length;
  variance = calcVar(parVals, mean);
  return;
  //## end Chain::calcStats%4198DB3602EE.body
}

Real Chain::calcVar (RealArray& parVals, Real mean) const
{
  //## begin Chain::calcVar%4198DC5C004B.body preserve=yes
  Real variance = .0;
  size_t sz = parVals.size();
  parVals -= mean;
  for (size_t i=0; i<sz; ++i)
  {
     Real parVals_i = parVals[i];
     variance += parVals_i*parVals_i;
  }
  return variance;
  //## end Chain::calcVar%4198DC5C004B.body
}

void Chain::findParsInChain (std::vector<Chain::ParamID>& parIDs, std::vector<size_t>& found) const
{
  //## begin Chain::findParsInChain%41F155CD01E6.body preserve=yes
   // Find location of parIDs in m_paramIDs.  This is an O(N^2)
   // algorithm, but N > 10^2 is currently unimaginable.
   found.clear();
   size_t nPars = m_paramIDs.size();
   for (size_t i=0; i<parIDs.size(); ++i)
   {
      const string& modName = parIDs[i].modName;
      size_t index = parIDs[i].index;
      for (size_t j=0; j<nPars; ++j)
      {
         const ParamID& chainPar = m_paramIDs[j];
         if (chainPar.modName == modName && chainPar.index == index)
         {
            found.push_back(j);
            // Also, fill in the remainder of the calling function's
            // parID struct.  We assume it doesn't have this info
            // till now.
            parIDs[i].units = chainPar.units;
            parIDs[i].parName = chainPar.parName;
            break;
         }
      }
   }
  //## end Chain::findParsInChain%41F155CD01E6.body
}

std::vector<Chain::ParamID> Chain::findParsInChain (size_t par1, const string& modName1, size_t par2, const string& modName2) const
{
  //## begin Chain::findParsInChain%4213A31302E8.body preserve=yes
    //get parameter names from header of first chain file

    //pos is not needed in this case, except to compile
    std::vector<size_t> pos;
    std::vector<Chain::ParamID> paramIDs(2);

    Chain::ParamID 
	& xParamID = paramIDs[0], & yParamID = paramIDs[1];

    xParamID.modName = modName1;
    xParamID.index = par1;
    yParamID.modName = modName2;
    yParamID.index = par2;

    findParsInChain(paramIDs, pos);

    return paramIDs;
  //## end Chain::findParsInChain%4213A31302E8.body
}

bool Chain::checkForRepeat (const std::vector<Real>& prevPoint, const std::vector<Real>& point)
{
  //## begin Chain::checkForRepeat%46114DAC02A6.body preserve=yes
   // ASSUMES both input vectors are same size, no checking.
   bool isRepeat = true;
   const Real EPS = 1.0e-15;
   const size_t nPar = prevPoint.size();
   for (size_t i=0; i<nPar; ++i)
   {
      Real prev = prevPoint[i];
      Real curr = point[i];
      Real diff = std::fabs((curr-prev)/curr);
      if (diff >= EPS)
      {
         isRepeat = false;
         break;
      }
   }
   return isRepeat;
  //## end Chain::checkForRepeat%46114DAC02A6.body
}

int Chain::FWIDTH ()
{
  //## begin Chain::FWIDTH%461BE25B029F.body preserve=yes
   return ChainIO::FWIDTH();
  //## end Chain::FWIDTH%461BE25B029F.body
}

void Chain::readPointFromLine (size_t lineNum, RealArray& parVals) const
{
  //## begin Chain::readPointFromLine%461D09C00141.body preserve=yes
  m_IO->doReadPointFromLine(lineNum, parVals);
  //## end Chain::readPointFromLine%461D09C00141.body
}

int Chain::TEMPERPREC ()
{
  //## begin Chain::TEMPERPREC%46D582B901C5.body preserve=yes
   return ChainIO::TEMPERPREC();
  //## end Chain::TEMPERPREC%46D582B901C5.body
}

const string& Chain::format () const
{
  //## begin Chain::format%46E848F203DE.body preserve=yes
   return m_IO->getFormatName();
  //## end Chain::format%46E848F203DE.body
}

// Additional Declarations
  //## begin Chain%4180176A01EC.declarations preserve=yes
  //## end Chain%4180176A01EC.declarations

//## begin module%418011BB031C.epilog preserve=yes
//## end module%418011BB031C.epilog
